
class Lines:
    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.points = []

    def add_point(self, x, y):
        self.points.append((x, y))

    def to_dict(self):
        return {
            'x': self.x,
            'y': self.y,
            'points': self.points
        }



class Pin:
    def __init__(self, name, number, function, description: str, x: int, y: int, length: int, rotation: int,
                 name_visible=True):
        self.name = name
        self.name_visible = name_visible
        self.number = number
        self.number_visible = True
        self.function = function
        self.description = description
        self.x = x
        self.y = y
        self.length = length
        self.rotation = rotation

    def to_dict(self):
        return {
            'name': self.name,
            'name_visible': self.name_visible,
            'number': self.number,
            'number_visible': self.number_visible,
            'function': self.function,
            'description': self.description,
            'x': self.x,
            'y': self.y,
            'length': self.length,
            'rotation': self.rotation
        }

    def from_dict(self, data):
        self.name = data['name']
        self.name_visible = data['name_visible']
        self.number = data['number']
        self.number_visible = data['number_visible']
        self.function = data['function']
        self.description = data['description']
        self.x = data['x']
        self.y = data['y']
        self.length = data['length']
        self.rotation = data['rotation']



import drawsvg as svg


def draw_circle(circle, style):
    return svg.Circle(circle.center_x, circle.center_y * -1, circle.radius, **style)


def draw_lines(lines, close, style):
    points = [lines.x, lines.y]
    for point in lines.points:
        points += list(point)
    return svg.Lines(*points,
                     close=close,
                     **style)


def draw_rectangle(rectangle, style):
    return svg.Rectangle(rectangle.x,
                         rectangle.y * -1,
                         rectangle.width,
                         rectangle.height,
                         **style)





class Footprint:
    def __init__(self, name):
        self.name = name
        self.width = 0
        self.height = 0
        self.pads = []
        self.top_overlay = []
        self.bottom_overlay = []
        self.assembly_top = []

    def add_pad(self, pad):
        if isinstance(pad, list):
            self.pads += pad
        else:
            self.pads.append(pad)

    def add_overlay(self, overlay):
        if isinstance(overlay, list):
            self.top_overlay += overlay
        else:
            self.top_overlay.append(overlay)



from src.part_library_gen.generators.components.symbol import Symbol, Part
from src.part_library_gen.generators.components.designator import Designator
from src.part_library_gen.generators.components.part_number import PartNumber
from src.part_library_gen.generators.components.pin import Pin
from src.part_library_gen.generators.components.line import Line
from src.part_library_gen.generators.components.lines import Lines
from src.part_library_gen.styles import default_style


def opamp_generator(component_data, generator_data):
    symbol = Symbol(part_number=component_data['part'],
                    designator=component_data['designator'])

    height = default_style.pin_spacing * 4
    width = default_style.pin_spacing * 5
    #symbol.width = width + 2 * 100 + 40
    #symbol.height = height + 120

    body_x = -width / 2
    body_y = -height / 2
    part = Part(PartNumber(text=component_data['part'], x=-50, y=body_y + height + 40),
                Designator(designator=component_data['designator'], x=-50, y=body_y - 10))
    part.width = width + 2 * 100 + 40
    part.height = height + 120

    # draw triangle body
    triangle = Lines(default_style.pin_spacing * 3, 0)
    triangle.add_point(0, default_style.pin_spacing * 2)
    triangle.add_point(0, default_style.pin_spacing * -2)
    triangle.add_point(default_style.pin_spacing * 3, 0)
    part.add_body(triangle)
    # draw '+' and '-' inside body
    plus1 = Line(x1=10, y1=default_style.pin_spacing, x2=30, y2=default_style.pin_spacing)
    plus2 = Line(x1=10 + 10, y1=default_style.pin_spacing + 10, x2=10 + 10, y2=default_style.pin_spacing - 10)
    minus = Line(x1=10, y1=default_style.pin_spacing * -1, x2=30, y2=default_style.pin_spacing * -1)
    part.add_body(plus1)
    part.add_body(plus2)
    part.add_body(minus)

    pins = find_pins(component_data)
    # add positive pin
    pos_pin_dict = component_data['pins'][pins['pos']]
    pos_pin = Pin(name=pins['pos'],
                  number=pos_pin_dict['no'],
                  function=pos_pin_dict['func'],
                  description=pos_pin_dict['desc'],
                  x=default_style.pin_length * -1,
                  y=default_style.pin_spacing,
                  length=default_style.pin_length,
                  rotation=0,
                  name_visible=False)
    part.add_pin(pos_pin)

    # add negative pin
    neg_pin_dict = component_data['pins'][pins['neg']]
    neg_pin = Pin(name=pins['neg'],
                  number=neg_pin_dict['no'],
                  function=neg_pin_dict['func'],
                  description=neg_pin_dict["desc"],
                  x=default_style.pin_length * -1,
                  y=default_style.pin_spacing * -1,
                  length=default_style.pin_length,
                  rotation=0,
                  name_visible=False)
    part.add_pin(neg_pin)

    # add output pin
    out_pin_dict = component_data['pins'][pins['out']]
    out_pin = Pin(name=pins['out'],
                  number=out_pin_dict['no'],
                  function=out_pin_dict['func'],
                  description=out_pin_dict["desc"],
                  x=default_style.pin_spacing * 3,
                  y=0,
                  length=default_style.pin_length,
                  rotation=0,
                  name_visible=False)
    part.add_pin(out_pin)

    if 'pwr_pos' in pins and 'pwr_neg' in pins:
        # add power pins
        pwr_pos_pin_dict = component_data['pins'][pins['pwr_pos']]
        pwr_pos_pin = Pin(name=pins['pwr_pos'],
                          number=pwr_pos_pin_dict['no'],
                          function=pwr_pos_pin_dict['func'],
                          description=pwr_pos_pin_dict["desc"],
                          x=default_style.pin_spacing,
                          y=default_style.pin_spacing * -3,
                          length=default_style.pin_length - 16,
                          rotation=90,
                          name_visible=False)
        part.add_pin(pwr_pos_pin)

        # add negative power pin
        pwr_neg_pin_dict = component_data['pins'][pins['pwr_neg']]
        pwr_neg_pin = Pin(name=pins['pwr_neg'],
                          number=pwr_neg_pin_dict['no'],
                          function=pwr_neg_pin_dict['func'],
                          description=pwr_neg_pin_dict["desc"],
                          x=default_style.pin_spacing,
                          y=default_style.pin_spacing * 3,
                          length=default_style.pin_length - 16,
                          rotation=270,
                          name_visible=False)
        part.add_pin(pwr_neg_pin)

    symbol.add_part(part)
    return symbol


def find_pins(component_data):
    pins = {'pos': None,
            'neg': None,
            'out': None}
    for pin in component_data['pins']:
        if pin.startswith('IN') and '+' in pin and 'InAnalog' in component_data['pins'][pin]['func']:
            pins['pos'] = pin
        if pin.startswith('IN') and '-' in pin and 'InAnalog' in component_data['pins'][pin]['func']:
            pins['neg'] = pin
        if pin.startswith('OUT') and 'OutAnalog' in component_data['pins'][pin]['func']:
            pins['out'] = pin
        if 'PwrIn' in component_data['pins'][pin]['func']:
            if '+' in pin or pin == 'VCC+':
                pins['pwr_pos'] = pin
            if '-' in pin or pin == 'GND':
                pins['pwr_neg'] = pin
    return pins



from decimal import Decimal


class Pad:
    def __init__(self, number: int, x: Decimal, y: Decimal, width: Decimal, height: Decimal):
        self.number = number
        self.x = x  # pad center
        self.y = y  # pad center
        self.width = width
        self.height = height

    def __str__(self):
        return f"Pad {self.number} ({self.x} {self.y} {self.width} {self.height})"

    def __repr__(self):
        return self.__str__()


class Designator:
    def __init__(self, designator, x, y):
        self.designator = designator
        self.x = x
        self.y = y

    def to_dict(self):
        return {
            'designator': self.designator,
            'x': self.x,
            'y': self.y
        }

    def from_dict(self, designator_dict):
        self.designator = designator_dict['designator']
        self.x = designator_dict['x']
        self.y = designator_dict['y']



from decimal import Decimal
from .footprint import Footprint
from ..components.pad import Pad
from ..components.lines import Lines
from ..components.circle import Circle
from ..components.rectangle import Rectangle


class Parameters:
    def __init__(self, pin_count: int, pin_pitch, pad_width, pad_height, pad_row_spacing):
        self.pin_count = pin_count
        self.pin_pitch = pin_pitch
        self.pad_width = pad_width
        self.pad_height = pad_height
        self.pad_row_spacing = pad_row_spacing


def generate(package_parameters, name):
    row_spacing = package_parameters.E.get_available_max() + 2 * package_parameters.b.get_available_max() - package_parameters.L.get_available_max()
    overlay_d = package_parameters.D.get_available_max()
    overlay_e = package_parameters.E.get_available_max() - package_parameters.L.get_available_max() - Decimal('0.5')
    parameters = Parameters(pin_count=package_parameters.pin_count,
                            pin_pitch=package_parameters.e.get_available_max(),
                            pad_width=package_parameters.L.get_available_max(),
                            pad_height=package_parameters.b.get_available_max() + Decimal('0.06'),
                            pad_row_spacing=package_parameters.L.get_available_max() + 2*package_parameters.b.get_available_max()-package_parameters.L.get_available_max())

    pin_count_per_side = int(parameters.pin_count / 2)
    first_pad_y = (pin_count_per_side - 1) * parameters.pin_pitch / 2
    left_pads = add_vertical_pads(first=1,
                                  last=pin_count_per_side,
                                  pitch=parameters.pin_pitch,
                                  width=parameters.pad_width,
                                  height=parameters.pad_height,
                                  x=row_spacing / -2,
                                  y_offset=first_pad_y)
    right_pads = add_vertical_pads(first=parameters.pin_count,
                                   last=pin_count_per_side + 1,
                                   pitch=parameters.pin_pitch,
                                   width=parameters.pad_width,
                                   height=parameters.pad_height,
                                   x=row_spacing / 2,
                                   y_offset=first_pad_y)

    footprint = Footprint(name)
    footprint.add_pad(left_pads)
    footprint.add_pad(right_pads)

    # add overlay layer
    overlay_rectangle = Lines(overlay_e / -2, overlay_d / 2)
    overlay_rectangle.add_point(overlay_e / -2, overlay_d / -2)
    overlay_rectangle.add_point(overlay_e / 2, overlay_d / -2)
    overlay_rectangle.add_point(overlay_e / 2, overlay_d / 2)
    footprint.add_overlay(overlay_rectangle)
    radius = Decimal('0.5')
    spacing = Decimal('0.5')
    overlay_one_position_mark = Circle(overlay_e / -2 + radius + spacing, overlay_d / 2 - radius - spacing, radius)
    footprint.add_overlay(overlay_one_position_mark)

    # add assembly layer
    width = package_parameters.E1.get_available_max()
    height = package_parameters.D.get_available_max()
    pad_max_x = package_parameters.E.get_available_max()/2
    footprint.assembly_top.append(Rectangle(width=width, height=height, x=-width/2, y=height/2))
    for i in range(pin_count_per_side):
        pad_height = package_parameters.b.get_available_max()
        footprint.assembly_top.append(Rectangle(width=pad_max_x - width/2,
                                                height=pad_height,
                                                x=-pad_max_x,
                                                y=first_pad_y + pad_height/2 - i*parameters.pin_pitch))
        footprint.assembly_top.append(Rectangle(width=pad_max_x - width / 2,
                                                height=pad_height,
                                                x=width / 2,
                                                y=first_pad_y + pad_height/2 - i * parameters.pin_pitch))

    footprint.width = row_spacing + 3
    footprint.height = overlay_d + 2
    return footprint


def add_vertical_pads(first, last, pitch: Decimal, width: Decimal, height: Decimal, x, y_offset: Decimal):
    pads = []
    count = abs(last - first)
    for i in range(count + 1):
        pads.append(Pad(number=first + i if first < last else first - i,
                        x=x,
                        y=y_offset - Decimal(i) * pitch,
                        width=width,
                        height=height
                        ))
    return pads



from decimal import Decimal


class Dimension:
    def __init__(self, typ, min_, max_):
        self.typ = typ
        self.min_ = min_
        self.max_ = max_

    def get_available_max(self):
        if self.max_:
            return self.max_
        if self.typ:
            return self.typ
        if self.min_:
            return self.min_

    @staticmethod
    def from_str(str_dimension):
        dim = Dimension(None, None, None)
        if '~' in str_dimension:
            tmp = str_dimension.split('~')
            dim.min_ = Decimal(tmp[0].replace('mm', ''))
            dim.max_ = Decimal(tmp[1].replace('mm', ''))
        elif '±' in str_dimension:
            tmp = str_dimension.split('±')
            dim.typ = Decimal(tmp[0].replace('mm', ''))
            dim.min_ = dim.typ - Decimal(tmp[1].replace('mm', ''))
            dim.max_ = dim.typ + Decimal(tmp[1].replace('mm', ''))
        elif 'max.' in str_dimension:
            dim.max_ = Decimal(str_dimension.replace('max.', '').replace('mm', ''))
        elif 'min.' in str_dimension:
            dim.min_ = Decimal(str_dimension.replace('min.', '').replace('mm', ''))
        else:
            dim.typ = Decimal(str_dimension.replace('mm', ''))
        return dim



import unittest
import json
from src.part_library_gen.generators.symbol_generator.multipart_generator import multipart_generator


class TestRectangularComponentGenerator(unittest.TestCase):
    def setUp(self):
        with open('tests/data/test_multipart.json') as json_file:
            j = json.load(json_file)
            self.symbol_data = j[0]
            self.generator_data = j[0]['symbol_generator']['multipart']

    def test_symbol_generation(self):
        generated_symbol = multipart_generator(self.symbol_data, self.generator_data)
        for i, symbol_part in enumerate(generated_symbol.parts):
            self.assertEqual(symbol_part.designator.designator, "U?")
            self.assertEqual(symbol_part.part_number.text, "Test Part Name")
            self.assertEqual(len(symbol_part.pins), 3)

        self.assertEqual(generated_symbol.parts[0].pins[0].name, 'IN1+')
        self.assertEqual(generated_symbol.parts[0].pins[1].name, 'IN1-')
        self.assertEqual(generated_symbol.parts[0].pins[2].name, 'OUT1')

        self.assertEqual(generated_symbol.parts[1].pins[0].name, 'IN2+')
        self.assertEqual(generated_symbol.parts[1].pins[1].name, 'IN2-')
        self.assertEqual(generated_symbol.parts[1].pins[2].name, 'OUT2')

        self.assertEqual(generated_symbol.parts[0].pins[0].number, 3)
        self.assertEqual(generated_symbol.parts[0].pins[1].number, 2)
        self.assertEqual(generated_symbol.parts[0].pins[2].number, 1)

        self.assertEqual(generated_symbol.parts[1].pins[0].number, 5)
        self.assertEqual(generated_symbol.parts[1].pins[1].number, 6)
        self.assertEqual(generated_symbol.parts[1].pins[2].number, 7)

        self.assertEqual(generated_symbol.parts[0].pins[0].function, 'InAnalog')
        self.assertEqual(generated_symbol.parts[0].pins[1].function, 'InAnalog')
        self.assertEqual(generated_symbol.parts[0].pins[2].function, 'OutAnalog')



import drawsvg as svg
from decimal import Decimal
from .detail import draw_circle
from .detail import draw_lines, draw_rectangle

from ...generators.components.circle import Circle
from ...generators.components.lines import Lines
from ...generators.components.rectangle import Rectangle


def export(footprint, filename=None):
    d = svg.Drawing(footprint.width, footprint.height, origin='center')

    for pad in footprint.pads:
        parts = generate_footprint_pad(pad)
        d.append(parts[0])
        d.append(parts[1])

    for element in footprint.top_overlay:
        if isinstance(element, Lines):
            style = {'stroke_width': 0.15,
                     'fill': 'none',
                     'stroke': 'white'}
            d.append(draw_lines(element,
                                close=True,
                                style=style))
        elif isinstance(element, Circle):
            style = {'stroke_width': 0.15,
                     'fill': 'none',
                     'stroke': 'white'}
            d.append(draw_circle(element, style))

    for element in footprint.assembly_top:
        if isinstance(element, Lines):
            style = {'stroke_width': 0.15,
                     'fill': 'none',
                     'stroke': 'white'}
            d.append(draw_lines(element,
                                close=True,
                                style=style))
        elif isinstance(element, Circle):
            style = {'stroke_width': 0.15,
                     'fill': 'none',
                     'stroke': 'white'}
            d.append(draw_circle(element, style))
        elif isinstance(element, Rectangle):
            style = {'stroke_width': 0.02,
                     'fill': 'none',
                     'stroke': 'blue'}
            d.append(draw_rectangle(element, style))

    if filename:
        d.set_pixel_scale(30)
        d.save_svg(f"{filename}.svg")


def generate_footprint_pad(pad):
    center_x = pad.x
    center_y = pad.y
    rect = svg.Rectangle(center_x - pad.width / 2,
                         (center_y + pad.height / 2) * -1,
                         pad.width,
                         pad.height,
                         stroke_width=0,
                         fill='red',
                         stroke='red')
    text = svg.Text(f"{pad.number}",
                    pad.height,
                    center_x,
                    center_y * -1 + Decimal('0.1'),
                    center=True)
    return [rect, text]



from .generators.footprint_generator.soic import generate as generate_soic
from .generators.footprint_generator.tssop import generate as generate_tssop
from .generators.footprint_generator.vssop import generate as generate_vssop
from .packages.tssop import TSSOP
from .packages.dimension import Dimension


generator_map = {
    "SOIC": (generate_soic, TSSOP),
    "TSSOP": (generate_tssop, TSSOP),
    "VSSOP": (generate_vssop, TSSOP)
}


def generate(data):
    generator_name = data["generator"]
    generator = generator_map[generator_name][0]
    if not isinstance(data["data"], generator_map[generator_name][1]):
        data["data"] = parse_data(data["data"], generator_map[generator_name][1])
    footprint_name = f"{generator_name}-{data["data"].pin_count}"
    return generator(data["data"], footprint_name), footprint_name


def parse_data(data, parameters_class):
    package_data = parameters_class()
    for parameter_name in data:
        if parameter_name == "pin_count":
            package_data.pin_count = data[parameter_name]
        else:
            setattr(package_data, parameter_name, Dimension.from_str(data[parameter_name]))
    return package_data



from src.part_library_gen.generators.components.symbol import Symbol
from .generator_map import generator_map


def multipart_generator(component_data, generator_data):
    symbol = Symbol(component_data['part'], component_data['designator'])
    for part in generator_data:
        part_generator_name = generator_data[part]["generator_name"]
        part_generator_data = generator_data[part]["generator_data"]

        symbol_dict = {
            "designator": component_data['designator'],
            "manufacturer": component_data['manufacturer'],
            "part": component_data['part'],
            "pins": {key: component_data['pins'][key] for key in generator_data[part]['pins']}
        }
        generated_symbol = generator_map[part_generator_name](symbol_dict, part_generator_data)
        symbol.add_part(generated_symbol.parts[0])
    return symbol



class Circle:
    def __init__(self, center_x, center_y, radius):
        self.center_x = center_x
        self.center_y = center_y
        self.radius = radius

    def to_dict(self):
        return {
            'center_x': self.center_x,
            'center_y': self.center_y,
            'radius': self.radius
        }



class Rectangle:
    def __init__(self, width, height, x, y):
        self.width = width
        self.height = height
        self.x = x
        self.y = y

    def to_dict(self):
        return {
            'width': self.width,
            'height': self.height,
            'x': self.x,
            'y': self.y
        }



from .default_generator import default_generator
from .opamp_generator import opamp_generator
from .rectangular_symbol_generator import rectangular_symbol_generator


generator_map = {'default': default_generator,
                 'rectangle': rectangular_symbol_generator,
                 'opamp': opamp_generator
                 }



class PartNumber:
    def __init__(self, text, x, y):
        self.text = text
        self.x = x
        self.y = y

    def to_dict(self):
        return {
            'text': self.text,
            'x': self.x,
            'y': self.y
        }

    def from_dict(self, part_number_dict):
        self.text = part_number_dict['text']
        self.x = part_number_dict['x']
        self.y = part_number_dict['y']



import drawsvg as svg
from drawsvg import Circle

from .pin import generate_symbol_pin
from ...generators.components.rectangle import Rectangle
from ...generators.components.line import Line
from ...generators.components.lines import Lines


def export(symbol, filename=None):
    parts_svg = []
    for i, part in enumerate(symbol.parts):
        d = svg.Drawing(part.width, part.height, origin='center')

        for element in part.body:
            if isinstance(element, Rectangle):
                d.append(svg.Rectangle(element.x,
                                       element.y,
                                       element.width,
                                       element.height,
                                       stroke_width=3,
                                       fill='yellow',
                                       stroke='black'))
            if isinstance(element, Lines):
                points = [element.x, element.y]
                for point in element.points:
                    points += list(point)
                d.append(svg.Lines(*points,
                                   stroke_width=4,
                                   fill='yellow',
                                   stroke='black'))
            if isinstance(element, Line):
                d.append(svg.Line(element.x1, element.y1, element.x2, element.y2, stroke_width=4, stroke='black'))

        for pin in part.pins:
            d.append(generate_symbol_pin(pin))

        d.append(svg.Text(part.designator.designator,
                          40,
                          part.designator.x,
                          part.designator.y))

        d.append(svg.Text(part.part_number.text,
                          40,
                          part.part_number.x,
                          part.part_number.y))

        if filename:
            if len(symbol.parts) > 1:
                d.save_svg(f"{filename}_{i}.svg")
            else:
                d.save_svg(f"{filename}.svg")
        parts_svg.append(d.as_svg())
    return parts_svg



class TSSOP:
    def __init__(self, pin_count=None, A=None, A1=None, b=None, c=None, D=None, D2=None, E=None, E1=None, E2=None,
                 e=None, L=None):
        self.pin_count = pin_count
        self.A = A
        self.A1 = A1
        self.b = b
        self.c = c
        self.D = D
        self.D2 = D2
        self.E = E
        self.E1 = E1
        self.E2 = E2
        self.e = e
        self.L = L
